<?php

require_once('inc/system.php');
require_once('inc/include.php');
require_once('inc/db.php');
require_once('inc/cache/paths.php');
/**
  @class Cache
  @brief Class for cacheing folders and items.
  @author Martin Zlomek
  @date 08-02-2006
 */
class Cache extends DB {
	public $paths; //Works differently for IMAP / LocalPOP

	/**
	  @brief Create a connection to the cache.
	  @param $user User object of the cache owner.
	 */
	public function __construct($user)
	{
    $this->paths = new Paths($user->getDir(),$this);

		parent::__construct($this->paths->getCacheDir());


	}

	/**
	  @brief Get the list of account folders.
	  @param $account Account object of the account.
	  @param $class Class of the folder objects.
	  @return The list as an array of $class objects.
	 */
	public function getFolders($account, $class)
	{
	  
		$this->transaction();
		$stmt = $this->prepare('SELECT folder_id, name, rights, sync
								FROM folder
								WHERE account_id = ?');
		$stmt->execute(array($account->accountID));

    
		$folders = array();
		
		while ($row = $stmt->fetch()) {
      $folder = new $class($account,
								 intval($row['folder_id']),
								 $row['name'],
								 intval($row['rights']),
                 $row['sync']);
                 
			$folders[$folder->name] = $folder;
		}
		
		$this->commit();
		return $folders;
	}
	
	/**
	  @brief Get the list of direct folder childern.
	  @param $account Account object of the account.
	  @param $class Class of the folder objects.
	  @param $folderID Id of the parent folder
	  @return The list as an array of $class objects.
	 */
	public function getChildern($account, $class, $folderID)
	{
		$this->transaction();
		$stmt = $this->prepare('SELECT folder_id, name, rights, sync
								FROM folder
								WHERE account_id = ? AND parent_folder_id = ?');
		$stmt->execute(array($account->accountID,$folderID));

		$folders = array();
		while ($row = $stmt->fetch()) {
			$folder = new $class($account,
								 intval($row['folder_id']),
								 $row['name'],
								 intval($row['rights']),
                 $row['sync']);
			$folders[$folder->name] = $folder;
		}
		$this->commit();

		if(count($folders)!=0) return $folders;
		
    return false;
	}
	
	/**
	  @brief Get the list of folder childern.
	  @param $account Account object of the account.
	  @param $class Class of the folder objects.
	  @param $folderID Id of the parent folder
	  @return The list as an array of $class objects.
	 */
	public function getParent($account, $class, $folderID)
	{
		$this->transaction();
		$stmt = $this->prepare('SELECT name, parent_folder_id
								FROM folder
								WHERE folder_id = ? AND account_id = ?');
		$stmt->execute(array($folderID,$account->accountID));


    while($row = $stmt->fetch())
            $parentID = $row['parent_folder_id'];
		
    $stmt = $this->prepare('SELECT name
								FROM folder
								WHERE folder_id = ?');
								
		$stmt->execute(array($parentID));            
    
    while($row = $stmt->fetch())
            $parentName = $row['name'];   
    
    //Type = 'main' in order to get parent imap folder
    $type = 'main';

    if (isset($parentName))
    	$folder = $account->getFolder($parentName,$type);
    else
    	$folder = null;

		$this->commit();

    return $folder;
	}	
	
	
	/**
	  @brief Get the list of folder items.
	  @param $folder Folder object of the folder.
	  @param $class Class of the item objects.
	  @param $filter SQL WHERE contition.
	  @param $orderBy SQL ORDER BY condition.
	  @param $offset SQL OFFSET statement.
	  @param $limit SQL LIMIT statement.
	  @return The list as an array of $class objects.
	 */
	public function getItems($folder, $class, $filter, $orderBy, $offset, $limit,$fields='')
	{

		$this->transaction();
		if ($filter)
			$filter = ' AND ' . $filter;
		if ($orderBy)
			$orderBy = ' ORDER BY ' . $orderBy;
		if ($limit)
			$limit = ' LIMIT ' . $offset . ',' . $limit;

		//Allowed fields - can select only theese fields from database
		$aAllowedFields = array(0=>'item_id',
														1=>'rid',
														2=>'size',
														3=>'date',
														4=>'flags',
														5=>'header_from',
														6=>'header_to',
														7=>'subject',
														8=>'static_flags',
														9=>'priority',
														10=>'color',
														11=>'smime_status',
														12=>'has_attachment',
                            13=>'*');

		
		//If no fields is set, add
		if (strpos($fields,'*')===true) 
    {          
            $bFields = false;
            $sFields = '*';
		
    } else{ 
            $bFields = true;
            if (substr_count($fields,'item_id')==0) $fields.=',item_id';

    $sFields = '';
		foreach($aAllowedFields as $field)
          if (substr_count($fields,$field)>0 || !$bFields) 
                $sFields .= $field.',';

    $sFields = substr($sFields,0,strlen($sFields)-1);  
    }

    // ODBC MS SQL limit handling    
    if($limit && $_SESSION['DBTYPE'] == 'ODBC'){

      // Extract $limit info
      $limit = str_replace('LIMIT ', '', $limit);
      $limit = explode(',', $limit);

      $stmt = $this->prepare('SELECT TOP '.intval($limit[1]). ' '.$sFields.' FROM 
                  (SELECT TOP '.intval($limit[1]+$limit[0]).' '.$sFields.',ROW_NUMBER() OVER('.$orderBy.') as row
                  FROM item WHERE folder_id = ' . $folder->folderID . $filter .' '.$orderBy.') x
                  WHERE x.row BETWEEN '.intval($limit[0]+1).' and '.intval($limit[1]+$limit[0]));
    }else{
      // Classic SQL
      $stmt = $this->prepare('SELECT '.$sFields.'
  								FROM item
  								WHERE folder_id = ' . $folder->folderID . $filter .
  								$orderBy . 
  								$limit);
    }

		$stmt->execute();
  
		$items = array();
		while ($row = $stmt->fetch()) {

			$item = new $class($folder,
								isset($row['item_id']) ? intval($row['item_id']) : '', 
								isset($row['rid']) ? $row['rid'] : '',
								isset($row['size']) ? intval($row['size']) : '', 
								isset($row['date']) ? $row['date'] : '',
								isset($row['flags']) ? intval($row['flags']) : '',
								isset($row['header_from']) ? $row['header_from'] : '', 
								isset($row['header_to']) ? $row['header_to'] : '',
								isset($row['subject']) ? $row['subject'] : '',
								isset($row['static_flags']) ? intval($row['static_flags']) : '',
								isset($row['priority']) ? intval($row['priority']) : '',
								isset($row['color']) ? $row['color'] : null,
								isset($row['smime_status']) ? intval($row['smime_status']) : '',
								(isset($row['has_attachment']) && $row['has_attachment'] == 'T') ? true : false);
			$items[$item->itemID] = $item;
		}
		
		$this->commit();

		return $items;
	}

	/**
	  @brief Get a folder item.
	  @param $folder Folder object of the folder.
	  @param $class Class of the item object.
	  @param $itemID ID of the item.
	  @return $class object.
	 */
	public function getItem($folder, $class, $itemID,$initialized = false)
	{
		if (!$initialized) $this->transaction();

		$stmt = $this->prepare('SELECT rid, size, date, flags, header_from, header_to, 
								subject, static_flags, priority, color, smime_status, has_attachment
								FROM item
								WHERE item_id = ? AND folder_id = ?');
		$stmt->execute(array($itemID,$folder->folderID));
		$row = $stmt->fetch();
		$stmt->closeCursor();
		if (!$initialized) $this->commit();
		if (!$row)
			throw new Exc('Invalid item');
			//return false;

		return new $class($folder,
						  $itemID, 
              $row['rid'],
						  intval($row['size']), 
							$row['date'],
						  intval($row['flags']),
						  $row['header_from'], 
							$row['header_to'],
						  $row['subject'],
							intval($row['static_flags']),
							intval($row['priority']),
							$row['color'],
							intval($row['smime_status']),
							($row['has_attachment'] == 'T') ? true : false);
						  
	}

	/**
	  @brief Get number of folder items.
	  @param $folder Folder object of the folder.
	  @param $flags Flags of counted items.
	  @param $positive If false, negative value of Flags will be used
	  @return The number of folder items.
	 */
	public function countItems($folder, $flags = 0, $positive = true,$filter = '')
	{		

		$this->transaction();
		if ($flags) {
			$not = $positive ? '' : 'NOT';
			
			$stmt = $this->prepare('SELECT COUNT(*) AS count
									FROM item
									WHERE folder_id = ? 
									AND ' . $not . ' (flags & ? > 0)');
			$stmt->execute(array($folder->folderID, $flags));
		} else {
			$stmt = $this->prepare('SELECT COUNT(*) AS count
									FROM item 
									WHERE folder_id = ? '.
                  ($filter?' AND '.$filter:''));
			$stmt->execute(array($folder->folderID));
		}


		$row = $stmt->fetch();
		$stmt->closeCursor();
		$this->commit();

		if (!$row)
			throw new Exc('Cannot get number of folder items');

		return intval($row['count']);
	}

	/**
	  @brief Get the list of cached folders IDs.
	  @param $accountID ID of the account.
	  @return The list as an array of (folderName => folderID).
	 */
	public function getFoldersID($accountID)
	{
		$this->transaction();
		$stmt = $this->prepare('SELECT folder_id, name, rights
								FROM folder
								WHERE account_id = ?');
		$stmt->execute(array($accountID));

		$folders = array();
		while ($row = $stmt->fetch()) {
			$folders[$row['name']] = array('folder_id' => intval($row['folder_id']),
											'rights'	=> intval($row['rights']));
		}
		$this->commit();

		return $folders;
	}

	/**
	  @brief Create a new cached folder.
	  @param $fields Fields of the new folder record.
	  @return ID of the new folder record.
	 */
	public function createFolder($fields)
	{
	  if (strtolower($fields['name'])=='inbox')
	   $fields['sync'] = '1';
		
    return $this->create('folder', $fields);
	}
	
	/**
	  @brief Cleanup the database (purge empty space).
	 */
	public function cleanup() {
		
		if ($_SESSION['DBTYPE']=="SQLITE")
			$this->query('VACUUM');
	}
	/**
	  @brief Delete a cached folder including sub-folders and items.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $depth Recursion depth.
	 */
	public function deleteFolder($accountID, $folderID, $depth = 0)
	{
		$this->transaction();

		/* Get sub-folders. */
		$stmt = $this->prepare('SELECT folder_id
								FROM folder
								WHERE parent_folder_id = ?');
		$stmt->execute(array($folderID));
		$folders = $stmt->fetchAll();

		/* Delete sub-folders. */
		foreach ($folders as $row) {
			$this->deleteFolder($accountID, intval($row['folder_id']),
								$depth + 1);
		}
		if ($folders) {
			$stmt = $this->prepare('DELETE FROM folder
									WHERE parent_folder_id = ?');
			$stmt->execute(array($folderID));
		}

		/* Delete the folder itself. */
		if (!$depth) {
			$stmt = $this->prepare('DELETE FROM folder
									WHERE folder_id = ?');
			$stmt->execute(array($folderID));
		}
		$stmt = $this->prepare('DELETE FROM item
								WHERE folder_id = ?');
		$stmt->execute(array($folderID));

		$this->commit();

		if (!$depth)
			$this->deleteMessageDir($accountID, $folderID);
	}
	/* Delete items 
   * @brief: deletes items from cache
	 * @param: $sFolder - folder to use
	 * @param: $oItems - if not present deletes all items in folder
	 */      	 
	public function deleteItems($oFolder,$oItems = false)
	{
		$this->transaction();

		if ($oItems!==false){
        foreach($oItems as $oItem){
          $this->deleteItem($oFolder->account->accountID,$oFolder->folderID,$oItem->itemID,$this);
          $this->deleteMessage($oFolder->account->accountID,$oFolder->folderID,$oItem->itemID);
        }
    } else { 
		    $stmt = $this->prepare("DELETE FROM item WHERE folder_id = ?");
		    $stmt->execute(array($oFolder->folderID));
        $this->deleteMessageDir($oFolder->account->accountID,$oFolder->folderID,true);
    }
    
    $this->commit();
    
	}

	/* Move items 
   * @brief: Moves items from cache folder to another
	 * @param: $oFrom - source folder object
	 * @param: $oTo   - target folder object	 
	 * @param: $oItems - if not present moves all items from source to target
	 */    	
	public function moveItems($oFrom,$oTo,$oItems = false)
	{
  	$this->transaction();
  	
  	//Cache fix
    $cache['cache'] = $this;             		  
		
    if ($oItems){ 
      foreach($oItems as $oItem){
        
        //TODO : imap does not return new ID's - slow down
        
        //Cache for localpop    
        if ($oFrom->itemClassName=='LocalPOPItem') 
          $this->updateItem($oItem->itemID,array('folder_id'=>$oTo->folderID),$cache);
        //Cache for IMAP
        else{
          $this->deleteItem($oFrom->account->accountID, $oFrom->folderID, $oItem->itemID,$this);
          $this->deleteMessage($oFrom->account->accountID,$oFrom->folderID,$oItem->itemID);
        }
      }
      
    }else{
      if ($oFrom->itemClassName=='LocalPOPItem'){
    	 $this->updateItem($oItem->itemID,array('folder_id'=>$oTo->folderID),$cache);
			} else {
        $stmt = $this->prepare('DELETE FROM item
								WHERE folder_id = ?'); 
				$stmt->execute(array($oFrom->folderID));
      }
      $this->deleteMessageDir($oFrom->account->accountID,$oFrom->folderID,true);
    }
    
    $this->commit();
    $oTo->sync();
	}
	
	/* Copy items 
   * @brief: Copies items from cache folder to another
	 * @param: $oFrom - source folder object
	 * @param: $oTo   - target folder object	 
	 * @param: $oItems - if not present copies all items from source to target
	 */    	
	public function copyItems($oFrom,$oTo,$oItems = false,$aServerID = false)
	{   
	  /*
	   * Work with ID's of new message - big preformance optimalization for LocalPOP,
	   * because cache doesn't need to synchronize target folder - it only creates new
	   * item in cache with rid(Server ID) that server returns
	   */          	  
	  if ($oFrom->itemClassName=='LocalPOPItem' && $aServerID){
      $cache['cache'] = $this;
      foreach($oItems as $oItem)
         $this->createItem(array('folder_id' => $oTo->folderID,
  										 'rid'     => $aServerID[$oItem->rid],
  										 'size'    => $oItem->size,
  										 'date'    => $oItem->date,
  										 'flags'   => $oItem->flags,
  										 'header_from' => $oItem->from,
  										 'header_to'   => $oItem->to,
  										 'subject' =>$oItem->subject,
  										 'static_flags'	 => $oItem->staticFlags,
  										 'priority'	 => $oItem->priority,
  										 'color'	 => $oItem->color,
  										 'smime_status' => $oItem->sMimeStatus,
  										 'has_attachment' => $oItem->hasAttachments ? 'T' : 'F'),$cache);


    }else {
	   $oTo->sync();
	/*	
   * RID problem in imap : imap_mail_copy , imap_mail_move doesnot return ids of
   * new messages 
   * 
   */
    }  
	}
		
	/**
	  @brief Update a cached folder.
	  @param $folderID ID of the folder.
	  @param $fields Fields to be updated.
	 */
	public function updateFolder($folderID, $fields)
	{
		$this->update('folder', array('folder_id' => $folderID), $fields);
	}

	/**
	  @brief Rename a cached folder.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $newFolderName New name of the folder.
	  @param $depth Recursion depth.
	  @param $length Length of the old name of the folder.
	 */
	public function renameFolder($accountID, $folderID,
								 $newFolderName, $depth = 0, $length = 0)
	{
		$this->transaction();

		if (!$depth) {
			$stmt = $this->prepare('SELECT length(name) AS name_length
									FROM folder
									WHERE folder_id = ?');
			$stmt->execute(array($folderID));

			$row = $stmt->fetch();
			$stmt->closeCursor();

			if (!$row)
				throw new Exc('Cannot get length of the name of the folder');

			$length = intval($row['name_length']) + 1;
		}

		/* Get sub-folders. */
		$stmt = $this->prepare('SELECT folder_id
								FROM folder
								WHERE parent_folder_id = ?');
		$stmt->execute(array($folderID));
		$folders = $stmt->fetchAll();

		/* Rename sub-folders. */
		foreach ($folders as $row) {
			$this->renameFolder($accountID, intval($row['folder_id']),
								$newFolderName, $depth + 1, $length);
		}
		if ($folders) {
			$stmt = $this->prepare('UPDATE folder
									SET name = ? || substr(name, ?, length(name))
									WHERE parent_folder_id = ?');
			$stmt->execute(array($newFolderName, $length, $folderID));
		}

		/* Rename the folder itself. */
		if (!$depth) {
			$stmt = $this->prepare('UPDATE folder
									SET name = ? || substr(name, ?, length(name))
									WHERE folder_id = ?');
			$stmt->execute(array($newFolderName, $length, $folderID));
		}

		$this->commit();
	}

	/**
	  @brief Get the list of cached items IDs.
	  @param $folderID ID of the folder.
	  @return The list as an array of (itemUID => itemID).
	 */
	public function getItemsID($folderID,$transaction = true)
	{
		if ($transaction)
      $this->transaction();
		$stmt = $this->prepare('SELECT item_id, rid, flags
								FROM item
								WHERE folder_id = ?');
		$stmt->execute(array($folderID));

		$items = array();
		while ($row = $stmt->fetch()) {
			$items[$row['rid']] = array('item_id' => intval($row['item_id']),
												'flags'	=> intval($row['flags']));
		}
		if ($transaction)
		  $this->commit();

		return $items;
	}

	
	/**
	  @brief Create a new cached item.
	  @param $fields Fields of the new item record.
	  @return ID of the new item record.
	 */
	public function createItem($fields,$cache = array())
	{
		return $this->create('item', $fields, $cache);
	}

	/**
	  @brief Delete a cached item.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $itemID ID of the item.
	 */
	public function deleteItem($accountID, $folderID, $itemID,$cache = false)
	{
		if ($cache===false) $this->transaction();
		$stmt = $this->prepare('DELETE FROM item
								WHERE item_id = ?');
		$stmt->execute(array($itemID));
		
		if ($cache===false) $this->commit();

		$this->deleteMessage($accountID, $folderID, $itemID);
	}

	/**
	  @brief Update a cached item.
	  @param $itemID ID of the item.
	  @param $fields Fields to be updated.
	 */
	public function updateItem($itemID, $fields, $cache = array())
	{
		$this->update('item', array('item_id' => $itemID), $fields, $cache);
	}

	/**
	  @brief Create a new record in the cache.
	  @param $table Name of the table where the new record will be created.
	  @param $fields Fields of the new record.
	  @return ID of the new record.
	 */
	private function create($table, $fields, $cache = array())
	{

		$columns = ' (' . implode(',', array_keys($fields)) . ') ';
		$values  = '(' . implode(',', array_fill(0, count($fields), '?')) . ')';

     if (!$cache['cache']) 
      $this->transaction();

		$stmt = $this->prepare('INSERT INTO ' . $table . $columns .
								'VALUES ' . $values);

		$stmt->execute(array_values($fields));
		
		if (!$cache['cache'])
      $this->commit();

		return $this->getID();
	}

	/**
	  @brief Update a record in the cache.
	  @param $table Name of the table where the new record will be created.
	  @param $id ID of the record.
	  @param $fields Fields of the new record.
	  @return ID of the new record.
	 */
	private function update($table, $id, $fields, $cache = array())
	{
		$columns = implode('=?,', array_keys($fields)) . '=? ';
		$selector = implode('=? AND ', array_keys($id)) . '=?';

		if (!$cache['cache']) $this->transaction();
		$stmt = $this->prepare('UPDATE ' . $table . ' SET ' . $columns .
								'WHERE ' . $selector);
		$stmt->execute(array_merge(array_values($fields), array_values($id)));
		if (!$cache['cache']) $this->commit();
	}
	

  public function markFolderItems($mailbox,$flag)
  {
      $stmt = $this->prepare('UPDATE item
								SET flags = flags | ? WHERE folder_id = ?');
			$stmt->execute(array($flag,$mailbox));    
  }
  
  public function unmarkFolderItems($mailbox,$flag)
  {
      $stmt = $this->prepare('UPDATE item
								SET flags = flags & ? WHERE folder_id = ?');
			$stmt->execute(array($flag,$mailbox));    
  }
  
  

	/**
	  @brief Get a cached message.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $ID ID of the message item.
	 */
	public function getMessage($accountID, $folderID, $itemID)
	{
		$file = $this->getMessageFile($this->getMessageDir($accountID, $folderID), $itemID);
    if (!is_file($file))
			return false;
		$return =  file_get_contents($file);
		return $return;
	}

	/**
	  @brief Create a new cached message.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $itemID ID of the message item.
	  @param $message The message itself.
	 */
	public function createMessage($accountID, $folderID, $itemID, $message)
	{
		$dir = $this->getMessageDir($accountID, $folderID);
		System::mkdir($dir);

		$file = $this->getMessageFile($dir, $itemID);
 		file_put_contents($file, $message);
	}
	
	/**
	  @brief Edit a cached message.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $itemID ID of the message item.
	  @param $message The message itself.
	 */
	public function editMessage($accountID, $folderID, $itemID, $message)
	{
		$file = $this->getMessageFile($this->getMessageDir($accountID, $folderID),
									  $itemID);
		file_put_contents($file, $message);
	}

	/**
	  @brief Delete cached message.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $itemID ID of the message item.
	 */
	public function deleteMessage($accountID, $folderID, $itemID)
	{
		$file = $this->getMessageFile($this->getMessageDir($accountID, $folderID),
									  $itemID);

    if (is_file($file))
			unlink($file);
	}

	/**
	  @brief Move cached message to another folder.
	  @param $accountID ID of the account.
	  @param $folderID ID of the folder.
	  @param $itemID ID of the message item.
	  @param $dstFolderID ID of the destination folder.
	 */
	public function moveMessage($accountID, $folderID, $itemID, $dstFolderID)
	{
		$file = $this->getMessageFile($this->getMessageDir($accountID, $folderID),
									  $itemID);
		if (is_file($file)) {
			$dstDir = $this->getMessageDir($accountID, $dstFolderID);
			System::mkdir($dstDir);

			@rename($file, $this->getMessageFile($dstDir, $itemID));
		}
	}

	/**
	  @brief Get the name of the message file.
	  @return The name.
	 */
	private function getMessageFile($dir, $itemID)
	{
		return $this->paths->getMessageFile($dir, $itemID);
	}

	/**
	  @brief Get the name of the message directory.
	  @return The name.
	 */
	private function getMessageDir($accountID, $folderID,$empty = false)
	{
		return $this->paths->getMessageDir($accountID, $folderID, $empty);
 	}
	
	/**
	  @brief Get the path to the message file.
	  @return The path.
	 */
	public function getMessagePath($accountID, $folderName, $itemID)
	{
	  return $this->paths->getMessagePath($accountID,$folderName,$itemID);
	}

	/**
	  @brief Delete message directory including subdirectories.
	 */
	private function deleteMessageDir($accountID, $folderID,$empty = false)
	{
	  
		System::rmdir($this->getMessageDir($accountID, $folderID, $empty),$empty);
	}

}

?>
