<?php
require_once('inc/folder.php');
require_once('inc/cache/cache.php');

/**
  @class CacheFolder
  @brief Class representing a cached folder.
  @author Martin Zlomek
  @date 02-03-2006
 */
abstract class CacheFolder extends Folder {
	/** Parent account object. */
	public $account;

	/** Internal ID of the folder. */
	public $folderID;

	/**
	  @brief Create a new folder object.
	  @param $account Parent account object of the folder.
	  @param $folderID ID of the folder.
	  @param $name Name of the folder.
	  @param $rights Rights to the folder.
	 */
	protected function __construct(&$account, $folderID, $name, $rights, $sync)
	{

    $this->sync = $sync?true:false;
		$this->account	= &$account;
		$this->folderID	= $folderID;
		
    parent::__construct($name, $rights);
	}

	/**
	  @brief Create a new folder.
	  @param $account Parent account object of the folder.
	  @param $name Name of the folder.
	  @param $rights Rights to the folder.
	  @return ID of the folder.
	 */
	protected static function create($account, $name, $rights)
	{
		$parentFolderName = preg_replace('/\/[^\/]*$/', '', $name);
		
		if (isset($account->folders['main'][$parentFolderName]))
			$parentFolderID = $account->folders['main'][$parentFolderName]->folderID;
		else
			$parentFolderID = null;

		$cache = new Cache($account->user,$account->protocol);
		return $cache->createFolder(array('parent_folder_id' => $parentFolderID,
										                  'account_id'		=> $account->accountID,
										                  'name'			 => $name,
										                  'rights'			=> $rights));
	}

	/**
	  @brief Delete the folder.
	 */
	public function delete()
	{
		$account = &$this->account;

		$cache = new Cache($account->user);
		$cache->deleteFolder($account->accountID, $this->folderID);
	}
	
	public function deleteItems($oItems = false)
	{
	  $account = &$this->account;
		
    $cache = new Cache($account->user);
		$cache->deleteItems($this,$oItems);

	}

  public function moveItems($oFolder,$oItems = false)
  {
      $cache = new Cache($this->account->user);
      $cache->moveItems($this,$oFolder,$oItems);	  
  }
  
  public function copyItems($oFolder,$oItems = false,$aCopiedID = false)
  {
      $cache = new Cache($this->account->user);
      $cache->copyItems($this,$oFolder,$oItems,$aCopiedID);	  
  }
	/**
	  @brief Empty the folder.
	 */
	protected function deleteAllItems()
	{
		$account = &$this->account;

		$cache = new Cache($account->user,$account->protocol);
		$cache->emptyFolder($account->accountID,$this->folderID);
	}
    
	/**
	  @brief Rename the folder.
	  @param $newName New name of the folder.
	 */
	protected function rename($newName)
	{
		$account = $this->account;

		$cache = new Cache($account->user,$account->protocol);
		$cache->renameFolder($account->accountID, $this->folderID,
							 $newName);

		$this->name = $newName;
	}
	/**
	  @brief Rename the folder.
	  @param $newName New name of the folder.
	 */
	public function setSubscription($state)
	{
		$account = $this->account;

		$cache = new Cache($account->user,$account->protocol);
		
    $cache->updateFolder($this->folderID,array('sync'=>$state));

		$this->sync = $state?true:false;
	}	


	/**
	  @brief Get the list of folder items.
	  @param $class Class of the item objects.
	  @param $filter SQL WHERE contition.
	  @param $orderBy SQL ORDER BY condition.
	  @param $offset SQL OFFSET statement.
	  @param $limit SQL LIMIT statement.
	  @return The list as an array of $class objects.
	 */
	public function getItems($class, $filter, $orderBy, $offset, $limit,$fields ='')
	{
	  $account = &$this->account;

		$cache = new Cache($account->user,$account->protocol);
		return $cache->getItems($this, $class, $filter, $orderBy, $offset, $limit,$fields);
	}

	/**
	  @brief Get an item.
	  @param $class Class of the item object.
	  @param $itemID ID of the item.
	  @return $class object.
	 */
	public function getItem($class, $itemID,$cache = array())
	{
		$account = &$this->account;

		if (!isset($cache['cache'])){
			$cache = array();
			$cache['cache'] = new Cache($account->user);
			$fCache = false;
		}else $fCache = true;
		return $cache['cache']->getItem($this, $class, $itemID, $fCache);
	}




	/**
	  @brief Load direct childern of the folder
	  @param $class Class of the folder objects.
	 */	
	public function getChildern($account)
	{
	     $account = &$this->account;

       $cache = new Cache($account->user,$account->protocol);
	     return $cache->getChildern($account, 
                                  $account->folderClassName, 
                                  $this->folderID);
  }
  
	/**
	  @brief Load direct parent of the folder
	  @param $class Class of the folder objects.
	 */	
	public function getParent($account)
	{
	     $account = &$this->account;
		
       $cache = new Cache($account->user,$account->protocol);
	     return $cache->getParent($account,
                                $account->classPrefix.'Folder',
                                $this->folderID);
  }
	
	/**
	  @brief Load direct parent of the folder
	  @param $class Class of the folder objects.
	 */	
	protected function getGWFolder($name)
	{
		try {
		  if ($this->account->gwAccount)
			 $folder = $this->account->gwAccount->getFolder($name); 
			else
			 $folder = false;
		} catch (Exc $e) {
			$folder = false;
		}
		return $folder;
	}


	/**
	  @brief Get number of items.
	  @param $flags Flags of counted items.
	  @param $positive If false, negative value of Flags will be used
	  @return The number of items.
	 */
	public function countItems($flags = 0, $positive = true,$filter = "")
	{
		$account = &$this->account;
    $cache = new Cache($account->user);
		return $cache->countItems($this, $flags, $positive,$filter);
	}
	
	public function markItems($flag)
	{
    $account = &$this->account;
    $cache = new Cache($account->user);
    $cache->markFolderItems($this->folderID,Item::FLAG_SEEN);
  }
  
  public function unmarkItems($flag)
  {
    $account = &$this->account;
    $cache = new Cache($account->user);
    $cache->unmarkFolderItems($this->folderID,Item::FLAG_SEEN);
  }
  
/**
	  @brief Synchronize the item cache with the list of items.
	  @param $class Class of the item object.
	  @param $aItemsID The list of item id's.
	 */
	protected function sync($class, $aItemsID,$folder)
	{
//    $t = time();
//    $m = microtime();
    
		$account = $this->account;
		
		//Retrieve cached items
		$cache['cache'] = new Cache($account->user);	  
//$t1 = time();
//$m1 = microtime();
		$aCacheItemsID = $cache['cache']->getItemsId($this->folderID);
/*if ($class =="POP3Item"){
    }*/
//$time1 = time()-$t1+microtime()-$m1;
//log_buffer("CacheFolder->Sync :: cache->getItemsId() (time:".$time1.")",'DEBUG'); 

//$t_edit = time();
//$m_edit = microtime();	
		//Get instance of imap to retrieve item details
		if ($class=='IMAPItem') {
			$protocolItem = IMAP::instance($account);
		} else if ($class=='LocalPOPItem') {
			$protocolItem = LocalPOP::instance($account);
		} else if ($class =="POP3Item"){
      $protocolItem = POP3::instance($account);
    }

    
		if ($aItemsID){
     $cache['cache']->transaction(); 
     foreach ($aItemsID as $oItem) {      

    //Item exists in databas - check for flags if imap item  
			if (isset($aCacheItemsID[$oItem->uid])) {
				$itemID = $aCacheItemsID[$oItem->uid];
				if ($class=='IMAPItem'){
				  $flags = IMAPItem::decodeFlags($oItem);
          if (isset($flags)) {
					 // keep flags higher than 64 (they are not stored to IMAP, just kept in database)
					 $newFlags = $flags | ($itemID['flags'] & ~63);
					 if ($newFlags != $itemID['flags'])
						  $cache['cache']->updateItem($itemID['item_id'], array('flags' => $newFlags),$cache);	
				    }
				}
				//Remove item from cacheditems ( remaining items in array will be
				//deleted if they are no longer in mailbox)
				unset($aCacheItemsID[$oItem->uid]);
        
			} else {
				//Append item to synchronization queue
				$aToSyncItems[] = $oItem;  
			}
		}	
	  $cache['cache']->commit(); 
  }
//$time_edit = time()-$t_edit+microtime()-$m_edit;
//log_buffer("[" . time()+microtime(). "] CacheFolder->Sync :: updateFlags (time:".$time_edit.")",'DEBUG'); 

		
		
		//Do chunked synchronization
		if ($aToSyncItems){ 
	  
		  
//$t2 = time();
//$m2 = microtime();		  
		  $lower = $protocolItem->msgNumber($aToSyncItems[0]->uid,$aItemsID);//!
//$time2 = time()-$t2+microtime()-$m2;
//log_buffer("CacheFolder->Sync ::ProtocolItem->msgNumber (time:".$time2.")",'DEBUG'); 
		
//$t3 = time();
//$m3 = microtime();		    	
      $aChunkedItems = array_chunk($aToSyncItems,CHUNK_SIZE);
//$time3 = time()-$t3+microtime()-$m3;
//log_buffer("CacheFolder->Sync :: array_chunk (time:".$time3.")",'DEBUG'); 
	
//$t4 = time();
//$m4 = microtime();	      
      
      if ($aChunkedItems) foreach($aChunkedItems as $aSyncItems){

				$cache['cache']->transaction();
				$upper = (($lower+CHUNK_SIZE-1)>count($aItemsID))?count($aItemsID):($lower+CHUNK_SIZE-1);
  //        $t5 = time();
 //         $m5 = microtime();	

    		if ($class=='LocalPOPItem'){
          foreach($aSyncItems as $k=>$v)
						$aMailItems[$k] = $protocolItem->getItem($this->name,$v->uid);
				}else if ($class=='POP3Item'){
          foreach($aSyncItems as $k=>$v)
						$aMailItems[$k] = $protocolItem->getItemHeaders($this->name,$v->uid);          
        } else
					$aMailItems = $protocolItem->getItems($this->name,$lower.":".$upper);

  //        $time5 = time()-$t5+microtime()-$m5;
  //        log_buffer("CacheFolder->Sync :: GetChunkedItemDetails($lower:$upper) (time:".$time5.")",'DEBUG'); 

				foreach($aMailItems as $oSyncItem){
				  //$t_t = time();
          //$m_t = microtime();
					
          call_user_func(array($class, 'decode'), $oSyncItem);  

					//Check subject encoding and try to convert it to UTF-8
					$cache['cache']->createItem(array('folder_id' => $this->folderID,
  										 'rid'     => $oSyncItem->uid,
  										 'size'    => $oSyncItem->size,
  										 'date'    => $oSyncItem->date,
  										 'flags'   => $oSyncItem->flags?$oSyncItem->flags:0,
  										 'header_from' => $oSyncItem->from,
  										 'header_to'   => $oSyncItem->to,
  										 'subject' =>$oSyncItem->subject,
  										 'static_flags'	 => $oSyncItem->staticFlags,
  										 'priority'	 => $oSyncItem->priority,
  										 'color'	 => $oSyncItem->color,
  										 'smime_status' => $oSyncItem->sMimeStatus,
  										 'has_attachment' => $oSyncItem->hasAttachments ? 'T' : 'F'),
  										 $cache);
					unset($aCacheItemsID[$key]);
					//$ti = time()-$t_t+microtime()-$m_t;
					//log_buffer("Cache- : Create item $oSyncItem->uid time($ti)",'DEBUG');
				}
//				if (DEBUG)
//          log_buffer("CacheFolder->Sync :: Commit sequence($lower:$upper)",'DEBUG');
				$cache['cache']->commit();
				$lower = ($upper+1);
			}    

 //       $time4 = time()-$t4+microtime()-$m4;
 //       log_buffer("CacheFolder->Sync :: Create items in cache (time:".$time4.")",'DEBUG'); 

    }
		//remove from sync list in lpop
		$protocolItem->drop($this->name);
		
		// Delete items from the item cache. 
		if ($aCacheItemsID){

			$cache['cache']->transaction();
			foreach ($aCacheItemsID as $itemID)  {
				$cache['cache']->deleteItem($account->accountID, $folder->folderID,$itemID['item_id'],$cache);
			}
			$cache['cache']->commit();
		}

 //     $time = time()-$t+microtime()-$m;
// //     log_buffer("CacheFolder->Sync: Sync (time:".$time.")",'DEBUG');
 //     global $timek;
 //     log_buffer("CacheFolder->Sync :: LocalPOP->GetItemDetails (totaltime:".$timek.")",'DEBUG');    
  }
  
	
}

require_once('inc/imap/folder.php');
require_once('inc/localpop/folder.php');
require_once('inc/pop3/folder.php');

?>
