<?php

require_once('inc/system.php');

/**
  @class DB
  @brief Class for manipulating with the SQLite database.
  @author Martin Zlomek
  @date 08-02-2006
 */
class DB extends PDO {
	/** Transaction counter. */
	private $transaction;

	/** Rolback flag. */
	private $rollback;

	/**
	  @brief Create a connection to a database.
	  @param $database Database to connect to.
	  @param[out] $newDatabase True if a new database is created.
	 */
	public function __construct($location)
	{
	
	  // Text field for table structure
	  $textfield = 'text';

	switch($_SESSION['DBTYPE']){
		case 'MYSQL':
		case 'ODBC':

		      // Field type handling
		      if ($_SESSION['DBTYPE']=='ODBC') 
		      { 
		        $autoincrement = 'IDENTITY'; 
			$textfield = 'varchar(255)';
		      } 
		      else 
		        $autoincrement = 'AUTO_INCREMENT';

          try{
            parent::__construct($_SESSION['DBCONN'],$_SESSION['DBUSER'],$_SESSION['DBPASS']);

            $stmt = $this->prepare("SELECT item_value FROM wm_metadata");
            $stmt->execute();

            $row = $stmt->fetch(PDO::FETCH_NUM);
           
            if (!is_array($row))  $newDatabase = true;
            else        $newDatabase = false;
       
          }catch (PDOException $e){
            //Try to create the database using mysql functions
            throw new Exc($e->getMessage(),"db_auto_create");
          }
			break;
		case 'SQLITE':
    default:
		  $autoincrement = 'AUTOINCREMENT';

      $filename = str_replace("%WMUSERPATH%",$location,$_SESSION['PURECONN']);       

			$newDatabase = !is_file($filename);
			
      // Create database directory. 
			if ($newDatabase)
				System::mkdir($location);			

		    $sConn = 'sqlite:'.$filename;

        parent::__construct($sConn);

        //Set pragmas
        $this->query('PRAGMA synchronous = 0;');
	      $this->query('PRAGMA temp_store = 2;');
      break;
	 }



  $this->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
  $this->setAttribute(PDO::ATTR_STRINGIFY_FETCHES, false);
  $this->transaction = 0;

 
	 if ($newDatabase){
	 try{

      $this->transaction();
			$this->query('CREATE TABLE folder (' /*)*/ .
						 'folder_id INTEGER PRIMARY KEY '.$autoincrement.', ' .
						 'parent_folder_id INTEGER, ' .
						 'account_id '.$textfield.' NOT NULL, ' .
						 'name '.$textfield.' NOT NULL, ' .
						 'rights INTEGER NOT NULL DEFAULT 0, '.
             'sync CHAR(1))'); 
			$this->query('CREATE TABLE item (' /*)*/ .
						 'item_id INTEGER PRIMARY KEY '.$autoincrement.', ' .
						 'folder_id INTEGER NOT NULL, ' .
						 'rid VARCHAR(32) NOT NULL, ' .
						 'size INTEGER NOT NULL, ' .
						 'date VARCHAR(32) NOT NULL, ' .
						 'header_from '.$textfield.' NOT NULL, ' .
						 'header_to '.$textfield.' NOT NULL, ' .
						 'subject '.$textfield.' NOT NULL, ' .
						 'priority INTEGER NOT NULL DEFAULT 0, ' .
						 'flags INTEGER NOT NULL DEFAULT 0, ' .
						 'static_flags INTEGER NOT NULL DEFAULT 0, ' .
						 'smime_status INTEGER NOT NULL DEFAULT 0, ' .
						 'has_attachment CHAR(1) DEFAULT \'F\', ' .
						 'color CHAR(1) DEFAULT \'Z\', ' .
						 'completed_on VARCHAR(32))');
			$this->query('CREATE INDEX IDX_item_folder_id ON item (folder_id)');
			$this->query('CREATE INDEX IDX_item_date ON item (date)');
			$this->query('CREATE TABLE wm_metadata (' .
						 'item_key VARCHAR(128), ' .
						 'item_value VARCHAR(128))');
			$stmt = $this->prepare('INSERT INTO wm_metadata (item_key, item_value) ' .
						 'VALUES (?, ?)');
			$stmt->execute(array('db_version', strval(WEBMAIL_DB_VERSION)));
			
	    $this->commit();
     
			
			}catch(PDOException $e){
        throw new Exc($e->getMessage());
      }
    }

 
  }
  
  

	/**
	  @brief Begin a transaction.
	 */
	public function transaction()
	{
		if (!$this->transaction++) {
			parent::beginTransaction();
			$this->rollback = false;
		}
	}

	/**
	  @brief Commit a transaction.
	  @return True on success, false on failure.
	 */
	public function commit()
	{
		if (!--$this->transaction) {
			if ($this->rollback)
				parent::rollback();
			else
				parent::commit();
		}
	}

	/**
	  @brief Roll back a transaction.
	  @return True on success, false on failure.
	 */
	public function rollback()
	{
		if (!--$this->transaction)
			parent::rollBack();
		else
			$this->rollback = true;
	}

	/**
	  @brief Get the ID of the last inserted row.
	  @return The ID.
	 */
	public function getID()
	{
		// PDO lastInsertID support
		if ($_SESSION['DBTYPE']!='ODBC') 
			return intval(parent::lastInsertId());

		// ODBC - MS SQL @@IDENTITY usage
		else
		{
	            $stmt = $this->prepare("SELECT @@IDENTITY AS ID");
        	    $stmt->execute();
	            $row = $stmt->fetch(PDO::FETCH_ASSOC);
		    return intval($row['ID']);
		}
	}

	/**
	  @brief Prepare a statement for execution.
	  @param $statement SQL statement.
	  @return PDOStatement object.
	 */
	public function prepare($statement)
	{
	  global $aError;
		try{
      $stmt = parent::prepare($statement);
    }
    catch(Exception $e){
          throw new Exc($e->getMessage(),$e->getCode());
    }
   // $stmt->setFetchMode(PDO::FETCH_ASSOC);

		return $stmt;
	}

	/**
	  @brief Execute a statement.
	  @param $statement SQL statement.
	  @return The result as an array of rows of SELECT statement,
	  or number of affected rows of INSERT, UPDATE or DELETE statements.
	 */
	public function query($statement)
	{
		switch (substr($statement, 0, 6)) {
		case 'SELECT':
		case 'CREATE':
			return parent::query($statement);

		case 'INSERT':
		case 'UPDATE':
		case 'DELETE':
		case 'VACUUM':
		case 'PRAGMA':
			return parent::exec($statement);

		default:
			throw new Exception('Unknown statement type');
		}
	}
}

?>
