<?php

require_once('inc/item.php');

/**
  @class IMAP
  @brief Class for accessing IMAP accounts.
  @author Martin Zlomek
  @revision Pavel Perna
  @date 17-04-2007
 */
class IMAP {
	/** Resource representing an IMAP stream to the mailbox. */
	private $imap;

	/** Server part of the mailbox name. */
	private $server;

	/** Mailbox hierarchy delimiter. */
	private $delimiter;
	
	/** Array of objects of this class - only one instance should be available for each login details */
	public static $instance;
  private static $aOpenedMailboxes;
	/***********************/
	/* IMAP object methods */
	/***********************/

	/**
	  @brief Get instance of IMAP class
	  For each combination server/port/username/password there is an only one instance of this class allowed.
	  @param $server Name of the server where the mailbox is located.
	  @param $port Port where connect to on the server.
	  @param $username Name of the owner of the mailbox.
	  @param $password Password to the mailbox.
	  @return Instance of IMAP class
	 */
	public static function instance($account) 
  {
    $index = $account->accountID;
		if (!isset(self::$instance[$index])) {
        $imap = __CLASS__;
        self::$instance[$index] = new $imap($account->server, $account->port, $account->username, $account->password);
    }

    return self::$instance[$index];
  }
  
  /* Prevent users to clone the IMAP object */
  public function __clone()
  {
    throw new Exception('Clone is not allowed.');
  }

  public function msgNumber($uid)
  {
    return imap_msgno($this->imap,$uid);
  }
  
	
	/**
	  @brief Open an IMAP stream to a mailbox.
	  @param $server Name of the server where the mailbox is located.
	  @param $port Port where connect to on the server.
	  @param $username Name of the owner of the mailbox.
	  @param $password Password to the mailbox.
	 */
	private function __construct($server, $port, $username, $password)
	{
		/* FIXME: SSL flag? */
		$attr = '';
		if ($port == 993)
			$attr .= '/ssl/novalidate-cert';
 
		$port = $port ? (':' . $port) : '';
    if ($port == ':110') $port.='/pop3/notls';
    
		$this->server = '{' . $server . $port .  $attr . '}';
   
   //dmp(self::$instance);

		$this->imap = @imap_open($this->server, $username, $password);
		
		if ($this->imap === false)
			self::error('Cannot open the IMAP stream','imap_open');

		/* Get the mailbox hierarchy delimiter. */
		$list = imap_getmailboxes($this->imap, $this->server, 'INBOX');

    if (!is_array($list) || !$list)
			self::error('Cannot get the mailbox hierarchy delimiter','imap_delimiter_retrieval');
		
    $this->delimiter = $list[0]->delimiter;
    //FIXME - pop3
    //if(!$this->delimiter) $this->delimiter = '/';
    
    return true;

	}

	/**
	  @brief Close the IMAP stream.
	 */
	public function __destruct()
	{
		imap_close($this->imap, CL_EXPUNGE);
	  
  }

	/***********/
	/* Mailbox */
	/***********/

	/**
	  @brief Get the list of mailboxes.
	  @param $ownerID ID of owner account
	  @return The list as an array of objects.
	 */
	public function getMailboxes($ownerID = 'owner')
	{
		
    $list = imap_getmailboxes($this->imap, $this->server, '*');/*imap_getsubscribed*/
		
    if (!is_array($list))
			self::error('Cannot get the list of mailboxes','imap_getmailboxes');

		$mailboxes = array();
		foreach ($list as $mailbox) {
			$name = preg_replace(/*{*/ '/^{[^}]*}/', '', $mailbox->name);
      $acl  = @imap_getacl($this->imap, $name);

			if (isset($acl['owner'])) {
				$mailbox->rights = $acl['owner'];
			} else if (isset($acl[$ownerID])) {
				$mailbox->rights = $acl[$ownerID];
			} else {
				// This is temporary hack, because ACL doesn't work with public mailboxes.
				// Therefore a full access is signified and in case the user cannot actually
				// modify the folder, an error will be generated during modification.
				// TODO: this code should be removed once an IMAP PHP extending function
				// for getting proper access rights is written
				$mailbox->rights = 'ldwrsipc';
			}
			
			$mailbox->name	= $this->decode($name);
			$mailboxes[$mailbox->name] = $mailbox;
		}

		return $mailboxes;
	}

	/**
	  @brief Create a new mailbox.
	  @param $mailbox Name of the mailbox.
	  @param $ownerID ID of owner account
	  @return Rights of the new mailbox.
	 */
	public function createMailbox($mailbox, $ownerID = 'owner')
	{
		$mbox	= $this->encode($mailbox);
		$mailbox = $this->server . $mbox;

		if (!imap_createmailbox($this->imap, $mailbox))
			self::error('Cannot create the mailbox','imap_mailbox_create');
		if (!imap_subscribe($this->imap, $mailbox))
			self::error('Cannot subscribe new mailbox','imap_mailbox_subscribe');

		$acl = @imap_getacl($this->imap, $mbox);

		if (isset($acl['owner'])) {
			$result = $acl['owner'];
		} else {
			$result = $acl[$ownerID];
		}
		//Store mailbox as opened
		$this->lastMailbox = $mailbox;
		
		return $result;
	}

	/**
	  @brief Delete a mailbox.
	  @param $mailbox Name of the mailbox.
	 */
	public function deleteMailbox($mailbox)
	{
		$mailbox = $this->server . $this->encode($mailbox);
		if (!imap_deletemailbox($this->imap, $mailbox))
			self::error('Cannot delete the mailbox','imap_mailbox_delete');
		if (!imap_unsubscribe($this->imap, $mailbox))
			self::error('Cannot unsubscribe deleted mailbox','imap_mailbox_unsubscribe');
	}

	/**
	  @brief Rename a mailbox.
	  @param $oldMailbox Name of the mailbox.
	  @param $newMailbox New name of the mailbox.
	 */
	public function renameMailbox($oldMailbox, $newMailbox)
	{
		$oldMailbox = $this->server . $this->encode($oldMailbox);
		$newMailbox = $this->server . $this->encode($newMailbox);
		
		if (!imap_renamemailbox($this->imap, $oldMailbox, $newMailbox))
			self::error('Cannot rename the mailbox','imap_mailbox_rename');
		if (!imap_unsubscribe($this->imap, $oldMailbox))
			self::error('Cannot unsubscribe renamed mailbox','imap_mailbox_unsubscribe_renamed');
		if (!imap_subscribe($this->imap, $newMailbox))
			self::error('Cannot subscribe renamed mailbox','imap_mailbox_subscribe_renamed');
	}

	/**
	  @brief Delete multiple items from cache
	 */
	public function deleteItems($oFolder,$oItems = false)
	{
	  $result = true;
		$this->openMailbox($oFolder->name);

    if ($oItems){
      foreach($oItems as $oItem)
          $sequence[] = $oItem->rid;
      $sequence = implode(",",$sequence);
  		
      if (!imap_setflag_full($this->imap, $sequence, '\\Deleted',ST_UID)
  		 || !imap_expunge($this->imap))
  		{
  			self::error('Cannot delete the items','imap_mailbox_delete_items');
  		}
		} else {
  		$nrMessages = imap_check($this->imap)->Nmsgs;
  		
  		if (!imap_setflag_full($this->imap, '1:'.$nrMessages, '\\Deleted')
  					|| !imap_expunge($this->imap))
  		{
  			self::error('Cannot empty the folder','imap_mailbox_empty_folder');
  		}
		}
	}
	/**
	  @brief Move items from the IMAP folder to another one.
	  @param $from Name of the source mailbox.
	  @param $to Name of the destination mailbox.
	 */
	public function moveItems($from, $to, $items = false,$copy = false)
	{
		$this->openMailbox($from->name);
		$nrMessages = imap_check($this->imap)->Nmsgs;
      
    $target = $this->encode($to->name);
    
		if ($items!==false){
  		foreach($items as $item)
  		  $ids[] = $item->rid;
  		
  		$ids = implode(",",$ids);
      if ($copy){
        if (!imap_mail_copy($this->imap, $ids, $target,CP_UID) 
  			 || !imap_expunge($this->imap))
          self::error('Cannot copy items','imap_mailbox_copyitems');
	     }else{
        if (!imap_mail_move($this->imap, $ids, $target,CP_UID) 
  			 || !imap_expunge($this->imap))
          self::error('Cannot move items','imap_mailbox_moveitems');       
       }
		}else
      if ($copy){ 
  		if (!imap_mail_copy($this->imap, '1:'.$nrMessages, $target) 
  			|| !imap_expunge($this->imap))
  			self::error('Cannot copy items','imap_mailbox_copyitems');
  		}else{
  		if (!imap_mail_move($this->imap, '1:'.$nrMessages, $target) 
  			|| !imap_expunge($this->imap))
  			self::error('Cannot move items','imap_mailbox_moveitems');        
      }
  }

	public function countMailboxItems($mailbox)
	{
	  $this->openMailbox($mailbox);
    return imap_check($this->imap)->Nmsgs;
  }

	/**
	  @brief Open an IMAP mailbox.
	  @param $mailbox Name of the mailbox.
	  @param $encoded Mailbox name encoded flag.
	 */
	private function openMailbox($mailbox, $encoded = false)
	{

    if (!$encoded)
			$mailbox = $this->server . $this->encode($mailbox);

    if ($this->lastMailbox == $mailbox)
			return;
		
		$this->lastMailbox = $mailbox;
		
		if (!imap_reopen($this->imap, $mailbox))
			self::error('Cannot open the mailbox','imap_mailbox_open');
	}		
	
	/********/
	/* Item */
	/********/

	/**
	  @brief Get the list of items of the mailbox.
	  @param $mailbox Name of the mailbox.
	  @return The list as an array of objects.
	 */
	public function getItemsID($mailbox,$sequence = '1:*')
	{
    $this->openMailbox($mailbox);
        
		if ($num = imap_num_msg($this->imap)) {
		
			$list = $this->fetchFast($sequence);
		  
      if (!$list)
				self::error('Cannot get the list of messages','imap_mailbox_message_list');

			return $list;
		}

		return array();
	}
	
  public function drop()
	{
	
  }
	
  public function getItems($mailbox,$sequence = '1:*')
	{
    $this->openMailbox($mailbox);
        
		if ($num = imap_num_msg($this->imap)) {
		
			$list = $this->fetchOverview($sequence);

      if (!$list)
				self::error('Cannot get the list of messages','imap_mailbox_message_list');

			return $list;
		}

		return array();
	}	
	
	/**
	  @brief Fetch overview of specified email messages
	  @param $sequence IMAP Sequence identifying the messages to fetch overview.
	  @return	Structure containing overview.
	 */
	private function fetchOverview($sequence,$fUID = 0)
	{	  
		return @imap_fetch_overview($this->imap, $sequence,$fUID);
	}
	
	/**
	  @brief Fetch basic info of specified email messages
	  @param $sequence IMAP Sequence identifying the messages to fetch info.
	  @return	Structure containing basic info.
	 */
	private function fetchFast($sequence)
	{
		return @imap_fetch_fast($this->imap, $sequence);    
	}
	
	/**
	  @brief Fetch message structure
	  @param $uid UID of the message.
	  @return	Structure containing message info.
	 */
	private function fetchStructure($uid)
	{
		return @imap_fetchstructure($this->imap, $uid, FT_UID);
  }

	/**
	  @brief Fetch message headers
	  @param $uid UID of the message.
	  @param $parse If TRUE, the headers will be parsed.
	  @return	Structure containing message headers.
	 */
	private function fetchHeaders($uid, $parse = false)
	{	
    $headers = @imap_fetchheader($this->imap, $uid, FT_UID);

    if ($parse) {
			
      $tmp = new MailParse($headers);
			$h =  $tmp->getHeaders();
      		  
      return $h;
		} else {
			return $headers;
		}
  }
  
  public function getAllHeaders()
  {
    return @imap_headers($this->imap);
  }
	
	/**
	  @brief Get details of specified message of the mailbox.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @param $getHeaders If TRUE, message headers will be included in result too.
	  @return The object containing detials of a message.
	 */
	public function getItemDetails($mailbox, $uid, $getHeaders = true, $item = false)
	{
    $this->openMailbox($mailbox);

    $sequence = @imap_msgno($this->imap, $uid);
        
    if (!$item) {
  		$list = $this->fetchFast($sequence);

  		if (!$list)
  			self::error('Cannot get the message details','imap_message_details');
  		$result = reset($list);
    } else
      $result = $item;

		if ($getHeaders)
			$this->appendMessageHeaders($result);

		return $result;
	}
	
	/**
	  @brief Append message headers to message info object
	  @param $info Object with message info (obtained by fetchFast() function).
	  @param $headers Headers array of the message.
	 */
	private function appendMessageHeaders(&$info, $headers = array())
	{
		if (!$headers) {
			$uid = $info->uid;
			$headers = $this->fetchHeaders($uid, true);
		}
			
		// merge partial results	
		foreach ($headers as $key => $value) {
			$info->$key = $value;
		}
	}

	/***********/
	/* Message */
	/***********/

	/**
	  @brief Get a message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @return Raw data of the message.
	 */
	public function getMessage($mailbox, $uid)
	{
		if ($mailbox)
		 $this->openMailbox($mailbox);

		$headers = $this->fetchHeaders($uid);
		if ($headers === false) 
			self::error('Cannot get the message headers','imap_no_headers');

		$body = @imap_body($this->imap, $uid, FT_UID);
		if ($body === false)
			self::error('Cannot get the message body','imap_no_body');
		return $headers . $body;
	}

	/**
	  @brief Create a new message in a mailbox.
	  @param $mailbox The mailbox.
	  @param $message Raw data of the message.
	  @return The message item.
	 */
	public function createMessage($mailbox, $message)
	{
    $encoded_mailbox = $this->server . $this->encode($mailbox);
    
		/* Create the message. */
		if (!@imap_append($this->imap, $encoded_mailbox, $message))
			self::error('Cannot create the message','imap_create_message');
		
		if ($mailbox!="INBOX" && !@imap_reopen($this->imap, $encoded_mailbox))
			self::error('Cannot open the mailbox','imap_open_mailbox');
		
		$parser = &new MailParse($message);
		$headers = $parser->getHeaders(true);


		$result = $this->getNewMessage($headers['message-id'], $encoded_mailbox);
		$this->appendMessageHeaders($result, $headers);
		

		
		return $result;
	}

	/**
	  @brief Delete a message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	 */
	public function deleteMessage($mailbox, $uid)
	{
		$this->openMailbox($mailbox);

		if (!@imap_delete($this->imap, $uid, FT_UID))
			self::error('Cannot delete the message','imap_delete_message');
	}

	/**
	  @brief Move a message to another mailbox.
	  @param $srcMailbox Name of the source mailbox.
	  @param $uid UID of the message.
	  @param $dstMailbox Name of the destination mailbox.
	  @param $copy If TRUE, the message will be copied instead of moved
	  @return New UID of the message.
	 */
	public function moveMessage($srcMailbox, $uid, $dstMailbox, $copy = false)
	{
		$this->openMailbox($srcMailbox);
		
		$mbox = $this->encode($dstMailbox);
		$mailbox = $this->server . $mbox;

		$msgID = $this->fetchOverview(imap_msgno($this->imap, $uid),FT_UID);
		$msgID = $msgID[0]->message_id;

		/* Move the message. */
		if ($copy)
		{
			if (!@imap_mail_copy($this->imap, $uid, $mbox, CP_UID))
			{
				self::error('Cannot copy the message','imap_copy_message');
			}
		} else {
			if (!@imap_mail_move($this->imap, $uid, $mbox, CP_UID)
				|| !@imap_expunge($this->imap))
			{
				self::error('Cannot move the message','imap_move_message');
			}
		}
		//FIXME
		//if ($dstMailbox=='Trash')
		$this->openMailbox($mailbox,true);
		return $this->getNewMessage($msgID, $mailbox)->uid;
	}

	/**
	  @brief Set flags on the message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @param $flags The flags.
	 */
	public function setFlags($mailbox, $uid, $flags)
	{
		$this->openMailbox($mailbox);

		if (!@imap_setflag_full($this->imap, $uid, $flags, ST_UID))
			self::error('Cannot set flags on the message','imap_set_flags');
	}

	/**
	  @brief Unset flags on the message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @param $flags The flags.
	 */
	public function unsetFlags($mailbox, $uid, $flags)
	{
		$this->openMailbox($mailbox);

		if (!@imap_clearflag_full($this->imap, $uid, $flags, ST_UID))
			self::error('Cannot clear flags on the message');
	}
	
	/**
	  @brief Set flags on the message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @param $flags The flags.
	 */
	public function setFolderFlags($mailbox, $flags)
	{
		$this->openMailbox($mailbox);

    $flags = IMAPItem::encodeFlags($flags);

		if (!@imap_setflag_full($this->imap, '1:*', $flags))
			self::error('Cannot set flags on the message','imap_set_flags');
	}

	/**
	  @brief Unset flags on the message.
	  @param $mailbox Name of the mailbox.
	  @param $uid UID of the message.
	  @param $flags The flags.
	 */
	public function unsetFolderFlags($mailbox, $flags)
	{
		$this->openMailbox($mailbox);
		
    $flags = IMAPItem::encodeFlags($flags);
    
		if (!@imap_clearflag_full($this->imap, '1:*', $flags))
			self::error('Cannot clear flags on the messages in the folder');
	}
	
	public function search($mailbox, $query){
    $this->openMailbox($mailbox);
    
    return imap_search($this->imap, $query,SE_UID);

  }
	
	
	/**
	  @brief Get just right created message in a mailbox.
	  @param $sequence Posible interval where the message could be.
	  @param $message Raw data of the message.
	  @param $getHeaders If true, message headers will be included in the result.
	  @return The message item.
	 */
	private function getNewMessage($msgID, $mailbox)
	{
		/* Get all possible items. */
		$new = imap_search($this->imap, 'RECENT');
		
    if (!$new)
			self::error('New message not found in destination mailbox - RECENT','imap_new_message_not_found');
	
		/*$sequence = '';
		foreach ($new as $item)
			$sequence .= ($sequence?',':'') . $item;
		*/
		
		$sequence = $new[0].":*";
		
		$candidates = $this->fetchOverview($sequence);
		if($msgID){
      if (count($candidates) == 1) {
  		  $item = reset($candidates);
        if($item->message_id == $msgID)
  			 return $item;
  		} else {
  			foreach ($candidates as $item) {
  				if ($item->message_id == $msgID)
  					return $item;
  			}
  		}
		}
		// Message not found:
		self::error('New message not found in destination mailbox','imap_new_message_not_found');
	}

	/****************/
	/* Tool methods */
	/****************/

	/**
	  @brief Convert mailbox name to IMAP format.
	  @param $mailbox Name of the mailbox.
	  @return Converted mailbox name.
	 */
	private function encode($mailbox)
  {
    if ($this->delimiter != '/')
    $mailbox = str_replace('/', $this->delimiter, $mailbox);
    
    return mb_convert_encoding($mailbox, 'UTF7-IMAP', 'UTF-8'); 
  }

  /**
  @brief Convert mailbox name from IMAP format.
  @param $mailbox Name of the mailbox.
  @return Converted mailbox name.
  */
  private function decode($mailbox)
  {
    $mailbox = mb_convert_encoding($mailbox, 'UTF-8', 'UTF7-IMAP');
    if ($this->delimiter != '/')
    $mailbox = str_replace($this->delimiter, '/', $mailbox);  
    return $mailbox;
  }

	/**
	  @brief Throw an Exc exception with the error message.
	  @param $message If there is no error message from the IMAP driver,
	  use this one.
	  @throw Exc exception with the error message.
	 */
	private static function error($message,$id)
	{
		$errors = imap_errors();
		if ($errors)
			throw new Exc($errors[0],'imap_internal');
		else 
			throw new Exc($message,$id);
	}
}

?>
