<?php

/*
	Parsed mail structure:

	array(
	 'headers'	 => array(header_name => header_value),
	 'attachments' => array('part_id',
							'cid',
							'type',
							'name',
							'size'),
	 'plain_body'  => array(
						body_?,
						...
						array(				  // submail_?
						  'headers',
						  'attachments',
						  'plain_body',
						  'html_body',
						),
						...
					  ),
	 'html_body'	=> array( [like plain_body] ), 
	)
 */

require_once('inc/encode.php');
require_once('inc/phpmailer/class.phpmailer.php');
require_once('inc/smime.php');
require_once('inc/include.php');

/**
 * @class MailParse
 * @brief Class for parsing rfc822 and rfc2045 email messages.
 * @author Martin Zlomek
 * @revision Pavel Perna
 * @date 17-04-2007  
 */
class MailParse {
	/** Resource representing the mail. */
	public $mail;
	
	/** This variable is used as a storage of decoded SMime message, if the message
	 file consists of the only one SMime decrypted part	*/
	private $sMimeMessage;

	/** The message. */
	private $message;

	/** Parts of the message. */
	private $parts;

	/** IDs of the message. */
	private $ids;
	
	/**
	  @brief Create the parser for a message.
	  @param $message The message.
	  @param $ids Array of all IDs of the message item.
	  @param $decodeMime Bool value - if true, Mime encoded message parts will be 
		decoded during parsing.
		@param $attIdPrefix Prefix for the attachmend IDs. Do not set this variable,
		it is used only when calling this class recursively for SMime encoded parts
		from $this->decryptPart() method.
	 */
	public function __construct($message, $ids = array(), $decodeMime = false, $attIdPrefix = '')
	{
		$this->mail = mailparse_msg_create();
		if (!mailparse_msg_parse($this->mail, $message))
			throw new Exception('Cannot parse the mail');


		$this->message = $message;
		$this->parts	= mailparse_msg_get_structure($this->mail);

    $this->ids = $ids;
		$this->decodeMime = $decodeMime;
		$this->attIdPrefix = $attIdPrefix;
		$this->sMimeMessage = '';

	}

	/**
	  @brief Free memory of the parser.
	 */
	public function __destruct()
	{
		mailparse_msg_free($this->mail);
	}

	/**
	  @brief Send a message part to the browser.
	  @param $partID ID of the part.
	 */
	public function sendPart($partID)
	{
		if (!isset($this->parts[$partID])) {
			return $this->recursivelyProcessEncryptedAttachments('sendPart', $partID);
		}

		$part = mailparse_msg_get_part($this->mail, $this->parts[$partID]);
		$info = mailparse_msg_get_part_data($part);

    ob_start();
		@mailparse_msg_extract_part($part, $this->message);
		
		
		$fileName = self::getAttachmentName($info, $partID);
		
		$length = ob_get_length();
		
		
		/* Send attachment data to output. */
		Tools::sendFileHeaders($fileName, $length, $info['content-type']);
 		ob_end_flush();

  }
  
  public function getPart($partID,&$params = array())
  {
		if (!isset($this->parts[$partID])) {
			return $this->recursivelyProcessEncryptedAttachments('getPart', $partID, $params);
		}
          
    $part = mailparse_msg_get_part($this->mail, $this->parts[$partID]);
    $info = mailparse_msg_get_part_data($part);
    $params['info'] = $info;

    $start = 'starting-pos' . (isset($params['full']) ? '' : '-body');
    $end = 'ending-pos'  . (isset($params['full']) ? '' : '-body');
		$partData = substr($this->message,$info[$start],$info[$end]-$info[$start]);
    return $partData;    
  }

	/**
	  @param Parse the mail.
	  @param[in,out] $partID ID of the part to be parsed.
	  @return Mail The parsed mail structure as described at the beginning of this file.
	 */
	public function parse(&$partID = 0)
	{
		$mail = array('headers'	 => array(),
					  'attachments' => array(),
					  'plain_body'  => array(),
					  'html_body'	=> array());

		$this->parsePart($partID, $mail);

		/* Substitute cid URIs. */
		if ($this->ids) {
			$search  = array();
			$replace = array();
			foreach ($mail['attachments'] as $key => $attachment) {
				if ($attachment['cid'] === '')
					continue;

				$search[] = '/(<((?i)img)[^>]*((?i)src)[^=]*=[^"]*("?))(((?i)cid:)' . /*)*/
					preg_quote($attachment['cid'], '/') . ')(\4[^>]*>)/';
				$fullpath = $this->ids['account_id'].'/'.$this->ids['folder_id'].'/'.
					$this->ids['item_id'].'/'.$this->attIdPrefix.$attachment['part_id'];
				
				$url = 'http'.($_SERVER['HTTPS']=='ON'?'s':'').'://'.$_SERVER['HTTP_HOST'].str_replace(WEBMAIL_PHP,DOWNLOAD_PHP,$_SERVER['SCRIPT_NAME']);
        
        $replace[] = '\1' .
					$url . '?sid=' . encodeURL($this->ids['sid']) .
					'&amp;class=attachment' .
					'&amp;fullpath=' . encodeURL($fullpath) .
					'\7';

        //Check if 
        if(preg_match(current($search),$body))
				  unset($mail['attachments'][$key]);
				//If not found in e-mail set as default attachment
        else
				  unset($mail['attachments'][$key]['cid']);
				
			}
			if ($search) {
				foreach ($mail['html_body'] as &$body) {
					if (!is_string($body))
						continue;

					$body = preg_replace($search, $replace, $body);
				}
				unset($body);
			}
		}
		
		if ($this->decodeMime) {
			$this->checkSMimeParts($mail);
			if ($this->sMimeMessage)
				 return $this->sMimeMessage;
		}
		return $mail;
	}
	
	/**
	  @brief Recursively decrypt encrypted attachment (specified by partID) and
		call method to process the decrypted part.
	  @param $funcName Name of the method to be called
		@param $partID ID of the part.
		@param $params Optional parameters for the method to be called
	 */
	private function recursivelyProcessEncryptedAttachments($methodName, $partID, &$params = null)
	{
		// find out if recursively SMimem encoded attachment is required
		$needle = '.decrypt.';
		$delimiter = strpos($partID, $needle);
		if ($delimiter === false)
			throw new Exc('Cannot get the part');
		
		$parser = $this->decryptPart(substr($partID, 0, $delimiter));
		if ($parser === false)
			throw new Exc('Cannot get the part');
			
		return $parser->$methodName(substr($partID,$delimiter + strlen($needle)), $params);
	}
	
	/**
	  @brief Process all attachments of a message. If an attachment is SMime encoded
		part, decode it and process the parsing function for the decoded part. If there
		is only one encoded part in whole message, store the decoded structure into
		$this->sMimeMessage variable.
	  @param[in,out] $structure Message strucrure of the MailParse function.
	 */
	private function checkSMimeParts(&$structure)
	{
		if (isset($structure['attachments'])) {
			foreach ($structure['attachments'] as $key => $att) {
				// encrypted part found, process it recursively
				if ($att['type'] == 'application/x-pkcs7-mime') {
					
					$parser = $this->decryptPart($att['part_id'], $store);
					if ($parser !== false) {
						$part = $parser->parse();
						// recursively add all encrypted attachments with special ID notation
						if (isset($part['attachments'])) {
							foreach ($part['attachments'] as &$partAtt) {
								$aID = strval($att['part_id']) . '.decrypt.' . strval($partAtt['part_id']);
								$partAtt['part_id'] = $aID;
								$structure['attachments'][] = $partAtt;
							}
						}
						
						// handle special case - if there is no body of the message and the first
						// attachment is being processed (the message is whole encrypted), store
						// the parsed structure of the decoded message into $this->sMimeMessage 
						// variable which will then be made available for calling function
						if ($key == 0 && $this->attIdPrefix == '' && !$structure['plain_body'] && !$structure['html_body']) {
								$this->sMimeMessage = $part;
						}
					}
				}
			}
		}
	}
	
	/**
	  @brief Get MailParse object of SMime encoded message part.
	  @param $partID ID of the SMime encoded message part.
	  @return MailParse object if the part can be decoded, false otherwise.
	 */
	private function decryptPart($partID) {
		$prdka = array('full' => true);
		$encrypted = $this->getPart($partID, $prdka);
		if (SMime::decodeMessage($encrypted)) {
			$attIdPrefix = $this->attIdPrefix . strval($partID) . '.decrypt.';
			$parser = new MailParse($encrypted, $this->ids, $this->decodeMime, $attIdPrefix);
			return $parser;
		}
		return false;
	}

	/**
	  @param Get headers of a message.
	 */
	public function getHeaders($plain = false)
	{
		$part = mailparse_msg_get_part($this->mail, $this->parts[0]);
		$info = mailparse_msg_get_part_data($part);

		if ($plain)
			return $info['headers'];
		else
			return self::parseHeaders($info);
	}
	
	/**
	  @brief Detects whether a message structure contains attachments.
	  @param $structure Structure of the message
	  @return TRUE or FALSE
	 */
	public static function hasAttachments($structure)
	{
		// find out whether any attachment without CID exists
		foreach ($structure['attachments'] as $attachment) {
			if ($attachment['cid'] === '')
				return true;
		}
		return false;
	}
	
	/**
	  @brief Detects whether a message structure contains HTML body.
	  @param $structure Structure of the message
	  @return TRUE or FALSE
	 */
	public static function hasHTMLBody($structure)
	{
		// find out whether HTML body exists
		if (is_array($structure['html_body']) && count($structure['html_body']) > 0)
			return true;
		else
			return false;
	}
	
	/**
	  @brief Detects whether a message is SMime signed - if so, verifies the signature.
	  @param $structure Structure of the message
	  @param $message Message itself.
	  @return 1 - message is not subscribed, 2 - message is subscribed and valid,
	  	3 - messaeg is subscribed and subscription is not valid
	 */
	public static function sMimeStatus($structure, &$message)
	{
		if (!isset($structure['attachments']))
			return 1;
		
		foreach ($structure['attachments'] as $item) {
			if ($item['type'] == 'application/x-pkcs7-signature') {
				if (SMime::verify($message))
					return 2;
				else
					return 3;
			}
		}
		
		// no signature found, return 0
		return 1;
	}
	
	/**
	  @param Parse a message part.
	  @param[in,out] $partID ID of the part to be parsed.
	  @param[out] $mail The parsed mail structure as described at the beginning of this file.
	 */
	private function parsePart(&$partID, &$mail)
	{
		$part = mailparse_msg_get_part($this->mail, $this->parts[$partID]);
		$info = mailparse_msg_get_part_data($part);

		if (isset($mail['headers']))
			$mail['headers'] = self::parseHeaders($info);
      


		if (isset($info['content-disposition'])) {
			switch ($info['content-type']) {
			case 'multipart/mixed':
      case 'text/plain':
			case 'text/html':
				if ($info['content-disposition'] == 'inline')
					break;

				/* Fall through. */
			default:
 
				$mail['attachments'][] = $this->parseAttachment($part, $info, $partID);

        // Embedded message - make sure we have the html and text parts already set
        if ($info['content-type']=='message/rfc822')
          // No html body but text body found        
          if (!count($mail['html_body']) && count($mail['plain_body']))
            // First plain/text part convert to html
            $mail['html_body'][]  = nl2br(htmlspecialchars($multipart['plain_body'][0]));

        $partID++;
				return;
			} 
		}

		if (strpos($info['content-type'], 'multipart/') === 0) {
			$part = $this->parts[$partID] . '.';

			$multipart = array('attachments' => array(),
								'plain_body'  => array(),
								'html_body'	=> array());

			$partID++;
			while ($partID < count($this->parts)) {
				if (strpos($this->parts[$partID], $part) !== 0)
					break;

				$this->parsePart($partID, $multipart);
			}
			$mail['attachments'] = array_merge($mail['attachments'],
												$multipart['attachments']);
      
			switch (substr($info['content-type'], 10)) {
			case 'report':
				$mail['plain_body'] = array_merge($mail['plain_body'], $multipart['plain_body']);
	
				// First plain/text part convert to html
				$mail['html_body'][]  = nl2br(htmlspecialchars($multipart['plain_body'][0]));

				break;

			case 'alternative':
				if ($multipart['plain_body'])
					$mail['plain_body'][] = end($multipart['plain_body']);
				if ($multipart['html_body'])
					$mail['html_body'][] = end($multipart['html_body']);
				break;

			case 'mixed':
			default:
				$mail['plain_body'] = array_merge($mail['plain_body'],
												  $multipart['plain_body']);
				$mail['html_body']  = array_merge($mail['html_body'],
												  $multipart['html_body']);
				break;
			}
		} else {
			switch ($info['content-type']) {
			case 'text/plain':
				if (isset($info['content-name']))
					$mail['attachments'][] = $this->parseAttachment($part, $info, $partID);
				else
					$mail['plain_body'][] = $this->parseBody($part, $info);
				break;

			case 'text/html':
				if (isset($info['content-name']))
					$mail['attachments'][] = $this->parseAttachment($part, $info, $partID);
				else
					$mail['html_body'][] = $this->parseBody($part, $info);
				break;

			default:
				$mail['attachments'][] = $this->parseAttachment($part, $info, $partID);
				break;
			}
			$partID++;
		}
	}

	/**
	  @brief Parse headers of the message.
	  @param $info Information about the message part.
	  @return The list of headers as an array of (header_name => header_value).
	 */
	private static function parseHeaders($info)
	{
		if (!isset($info['headers']))
			return array();

		$headers = $info['headers'];
		
		// fill result with default values (for the case they are not presented in $headers)
		$result = array('date' => -1,
							'from' => '',
							'to' => '',
							'cc' => '',
							'bcc' => '',
							'reply-to' => '',
							'subject' => '',
							'priority' => Item::NORMAL_PRIORITY,
							'custom-headers' => array());
							
		foreach ($headers as $key => $value) {
			switch ($key) {
				case 'date':
				  $date = @intval(strtotime($value));
				  if ($date > 0)
					  $result['date'] = date('Y-m-d H:i:s', $date);
					break;
					
				case 'from':
				case 'to':
				case 'cc':
				case 'bcc':
				case 'reply-to':
				case 'subject':
				case 'message-id':
					$result[$key] = self::decodeMimeHeader($value);
					break;
				
				case 'x-priority':
					$result['priority'] = self::decodePriority($value);
					break;
				
				// these headers shouldn't be sent to the output
				case 'received':
				case 'content-type':
				case 'mime-version':
					break;
				
				default:
					$result['custom-headers'][$key] = self::decodeMimeHeader($value);
					break;
			}
		}

		return $result;
	}

	/**
	  @brief Parse information about an attachment.
	  @param $part The attachment part.
	  @param $info Information about the part.
	  @param $partID ID of the part.
	  @return Information about the attachment
	  as an array of ('part_id', 'cid', 'type', 'name', 'size').
	 */
	private function parseAttachment($part, $info, $partID)
	{
    $info = mailparse_msg_get_part_data($part);
    $size = $info['ending-pos-body']-$info['starting-pos-body'];

    if ($info['transfer-encoding']=='base64'){
      $size = $size *(3/4);
    }
    
   	return array(
			'part_id' => $partID,
			'cid'	 => isset($info['content-id']) ? $info['content-id'] : '',
			'type'	=> $info['content-type'],
			'name'	=> self::getAttachmentName($info, $partID),
			'size'	=> $size,
		);

	}

	/**
	  @brief Get the name of an attachment.
	  @param $info Information about the attachment part.
	 */
	private static function getAttachmentName($info, $partID)
	{
		if (isset($info['content-name']))
      $name = self::decodeMimeHeader($info['content-name']);
			 
		
    if (isset($info['disposition-filename']))
      $name = self::decodeMimeHeader($info['disposition-filename']);	

    if ($name)
      return $name;
    else
		  return $partID.'.eml';
	}

	/**
	  @brief Parse a body.
	  @param $part The body part.
	  @param $info Information about the part.
	  @return The body.
	 */
	private function parseBody($part, $info)
	{
    
		ob_start();

		@mailparse_msg_extract_part($part, $this->message."\r\n\r\n");
		$data = ob_get_contents();
		ob_end_clean();

		// If not UTF8 and extended ASCII
		if ((strtoupper($info['content-charset']) != 'UTF-8') && Tools::isabovechar($data))
		{
 	                $data = iconv($info['content-charset']?$info['content-charset']:$_SESSION['DEFAULTCHARSET'], 'UTF-8//IGNORE', $data);
		}
      
		return $data;
	}

	/**
	  @brief Decode mime header.
	  @param $header Header to be decoded.
	  @return Decoded header.
	 */
	public static function decodeMimeHeader($header)
	{
      $converted = '';
      foreach (imap_mime_header_decode($header) as $item) {

        if ($item->charset == 'default' || strtolower($item->charset) == 'us-ascii'){
          
           if (!Tools::isabovechar($item->text))
            $converted .= $item->text;
           else
            $converted .= iconv($_SESSION['DEFAULTCHARSET'], 'UTF-8//IGNORE', $item->text);
        }
        else{
          $converted .= iconv($item->charset, 'UTF-8//IGNORE', $item->text);
        }
      }
      return (string) $converted;
	}
	
	/**
	  @brief Decode message priority.
	  @param $priority Message priority.
	  @return Decoded priority
	 */
	public static function decodePriority($priority)
	{
		ereg('^([0-9]{1,})',$priority,$regs);
		@$tmp = intval(reset($regs));
		return $tmp ? $tmp : Item::NORMAL_PRIORITY;
	}
	
	/**
	  @brief Parse addresses from a header (to, from, cc).
	  @param $header Header to be parsed.
	  @return Array containing addresses of a header.
	 */
	public static function parseAddresses($header)
	{
		return mailparse_rfc822_parse_addresses($header);
	}
}

?>
